"""Data definitions for the ray cluster setup.
Defines how the .daft.toml file should be structured.

The primary entrypoint into this module is the `build_ray_config_from_path` function.
"""

import ray
import sys
import click
from dataclasses import dataclass, field
from typing import Literal, Optional, Union, Any, List
from pathlib import Path

if sys.version_info >= (3, 11):
    import tomllib
else:
    import tomli as tomllib


RayConfiguration = dict[str, Any]


def _determine_python_version() -> str:
    maj = sys.version_info.major
    min = sys.version_info.minor
    mic = sys.version_info.micro
    return f"{maj}.{min}.{mic}"


@dataclass
class Setup:
    name: str
    provider: Literal["aws"]
    python_version: str = field(default_factory=_determine_python_version)
    ray_version: str = field(default=ray.__version__)
    number_of_workers: int = field(default=2)
    dependencies: List[int] = field(default_factory=list)


@dataclass
class Run:
    pre_setup_commands: List[str] = field(default_factory=list)
    setup_commands: List[str] = field(default_factory=list)


@dataclass
class Configuration:
    setup: Setup
    run: Run = field(default_factory=Run)


@dataclass
class AwsConfiguration(Configuration):
    region: str = field(default="us-west-2")
    ssh_user: str = field(default="ec2-user")
    instance_type: str = field(default="m7g.medium")
    image_id: str = field(default="ami-01c3c55948a949a52")
    iam_instance_profile_arn: Optional[str] = None


def _generate_setup_commands(
    config: Configuration,
) -> List[str]:
    return [
        "curl -LsSf https://astral.sh/uv/install.sh | sh",
        f"uv python install {config.setup.python_version}",
        f"uv python pin {config.setup.python_version}",
        "uv venv",
        """echo "alias pip='uv pip'" >> $HOME/.bashrc""",
        'echo "source $HOME/.venv/bin/activate" >> $HOME/.bashrc',
        "source $HOME/.bashrc",
        f'uv pip install "ray[default]=={config.setup.ray_version}" "getdaft" "deltalake"',
    ]


def _construct_config_from_raw_dict(custom_config: dict[str, Any]) -> Configuration:
    if "setup" not in custom_config:
        raise click.UsageError("No setup section found in config file.")
    if "provider" not in custom_config["setup"]:
        raise click.UsageError(
            "No provider value found in the setup section in the config file."
        )
    provider = custom_config["setup"]["provider"]
    if provider == "aws":
        setup = Setup(**custom_config["setup"])
        run = Run(**custom_config["run"])
        return AwsConfiguration(setup=setup, run=run)
    else:
        raise click.UsageError(f"Cloud provider {provider} not found")


def _construct_config_from_path(custom_config_path: Path) -> Configuration:
    try:
        with open(custom_config_path, "rb") as stream:
            custom_config = tomllib.load(stream)
            return _construct_config_from_raw_dict(custom_config)
    except click.UsageError as ce:
        raise ce
    except TypeError as te:
        (arg,) = te.args
        error = str(arg).removeprefix("Setup.__init__() g")
        raise click.UsageError(f"G{error}")
    except Exception as arg:
        raise click.UsageError(f"Error reading config file {custom_config_path}")


def _build_ray_config(
    custom_config: Configuration,
) -> RayConfiguration:
    if custom_config.setup.provider == "aws":
        aws_custom_config: AwsConfiguration = custom_config  # type: ignore
        return {
            "cluster_name": aws_custom_config.setup.name,
            "provider": {
                "type": "aws",
                "region": aws_custom_config.region,
                "cache_stopped_nodes": False,
            },
            "auth": {
                "ssh_user": aws_custom_config.ssh_user,
            },
            "max_workers": aws_custom_config.setup.number_of_workers,
            "available_node_types": {
                "ray.head.default": {
                    "resources": {},
                    "node_config": {
                        "InstanceType": aws_custom_config.instance_type,
                        "ImageId": aws_custom_config.image_id,
                    },
                },
                "ray.worker.default": {
                    "resources": {},
                    "node_config": {
                        "InstanceType": aws_custom_config.instance_type,
                        "ImageId": aws_custom_config.image_id,
                    },
                    "min_workers": aws_custom_config.setup.number_of_workers,
                    "max_workers": aws_custom_config.setup.number_of_workers,
                },
            },
            "setup_commands": aws_custom_config.run.pre_setup_commands
            + _generate_setup_commands(aws_custom_config)
            + aws_custom_config.run.setup_commands,
        }
    else:
        raise Exception("unreachable")


def build_ray_config_from_path(custom_config_path: Path) -> RayConfiguration:
    """Takes in a path to a file and returns a RayConfiguration object.

    # Assumptions:
    Assumes the path is a valid path to a file that exists.
    If it does not, the error printed to the console will be slightly misleading.
    Please check for existence beforehand.
    """

    custom_config = _construct_config_from_path(custom_config_path)
    return _build_ray_config(custom_config)



import asyncio
from botocore.exceptions import TokenRetrievalError
from typing import List, Optional, Any
from pathlib import Path
import subprocess
from . import helpers, data_definitions
from ray.autoscaler import sdk as ray_sdk
from ray import job_submission
import click
import time


AWS_TEMPLATE_PATH = Path(__file__).parent / "assets" / "aws.toml"
ON_CONNECTION_MESSAGE = """Successfully connected to Ray Cluster!

To view your cluster dashboard, navigate to: http://localhost:8265.

To run Daft against your Ray cluster, use the following code snippet:

```
import daft

daft.context.set_runner_ray(address="ray://localhost:10001")

df = daft.from_pydict({"foo": [1, 2, 3], "bar": [4, 5, 6]})
df.show()
```

Happy daft-ing! 🚀"""


def init_config(name: Path):
    with open(AWS_TEMPLATE_PATH) as template_f:
        with open(name, "w") as config_f:
            config_f.write(template_f.read())
            print(f"Successfully created a new configuration file: {name}")


def up(ray_config: data_definitions.RayConfiguration):
    ray_sdk.create_or_update_cluster(
        ray_config,
        no_restart=False,
        restart_only=False,
        no_config_cache=True,
    )
    print(f"Head node IP: {ray_sdk.get_head_node_ip(ray_config)}")
    print("Successfully spun the cluster up.")


def list():
    state_map = helpers.list_helper()
    for state_index, (state, instance_infos) in enumerate(state_map.items()):
        if state_index != 0:
            print()
        print(f"{state.replace('-', ' ').capitalize()}:")
        for index, instance_info in enumerate(instance_infos):
            if index != 0:
                print()
            print(str(instance_info))


def connect(
    ray_config: data_definitions.RayConfiguration,
    identity_file: Optional[Path],
):
    if not identity_file:
        identity_file = helpers.detect_keypair(ray_config)
    process = helpers.ssh_helper(ray_config, identity_file, [10001])
    print(ON_CONNECTION_MESSAGE)
    process.wait()


def submit(
    ray_config: data_definitions.RayConfiguration,
    identity_file: Optional[Path],
    working_dir: Path,
    cmd_args: List[str],
):
    if not identity_file:
        identity_file = helpers.detect_keypair(ray_config)
    cmd = " ".join(cmd_args)

    process = helpers.ssh_helper(ray_config, identity_file)
    try:
        working_dir_path = Path(working_dir).absolute()
        client = None
        tries = 0
        max_tries = 10
        while tries <= max_tries:
            try:
                client = job_submission.JobSubmissionClient(
                    address="http://localhost:8265"
                )
                break
            except Exception as e:
                tries += 1
                if tries >= max_tries:
                    raise e
                time.sleep(0.1)
        assert client
        job_id = client.submit_job(
            entrypoint=cmd,
            runtime_env={
                "working_dir": working_dir_path.absolute(),
            }
            if working_dir
            else None,
        )
        print(f"Job ID: {job_id}")

        asyncio.run(helpers.wait_on_job(client.tail_job_logs(job_id)))
        status = client.get_job_status(job_id)
        assert status.is_terminal(), "Job should have terminated"
        job_info = client.get_job_info(job_id)
        print(f"Job completed with {status}")

    finally:
        process.terminate()


def sql(
    ray_config: data_definitions.RayConfiguration,
    identity_file: Optional[Path],
    cmd_args: List[str],
):
    submit(
        ray_config,
        identity_file,
        Path(__file__).parent / "assets",
        ["python", "sql.py"] + cmd_args,
    )


def down(ray_config: data_definitions.RayConfiguration):
    ray_sdk.teardown_cluster(ray_config)
    print("Successfully spun the cluster down.")



import argparse
import daft


def main(sql: str):
    daft.context.set_runner_ray()
    df = daft.sql(sql).collect()
    df.show()


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("sql")
    args = parser.parse_args()
    main(args.sql)



"""Generic helpers.

Contain helpers for sshing, parsing outputs from `aws` commands, etc.

# Note
All helper/utility functions should go in here.
All core functions should go elsewhere.
"""

import asyncio
from dataclasses import dataclass, field
from typing import List, Literal, Optional, Any, Union
from pathlib import Path
import subprocess
import json
import click


@dataclass
class Tag:
    Key: str
    Value: str


@dataclass
class InstanceInformation:
    instance_id: str
    iam_role: str
    state: Union[
        Literal["terminated"], Literal["shutting-down"], Literal["running"], Any
    ]
    ip: str
    tags: List[Tag]

    @staticmethod
    def from_dict(data: dict) -> "InstanceInformation":
        tags_data = data.get("tags", [])
        tags = [
            Tag(**tag) for tag in tags_data
        ]  # Convert each dict in 'tags' to a Tag instance
        instance_info = InstanceInformation(**data)
        instance_info.tags = tags
        return instance_info

    def __str__(self) -> str:
        name = self.get_tag("ray-cluster-name")
        node_type = self.get_tag("ray-node-type")
        return f"""\tName: {name} ({node_type})
        Instance ID: {self.instance_id}
        IAM Role: {self.iam_role}
        State: {self.state}
        Ip: {self.ip}"""

    def get_tag(self, tag_name: str) -> Optional[str]:
        for tag in self.tags:
            if tag.Key == tag_name:
                return tag.Value


def _parse_describe_instances_query() -> List[InstanceInformation]:
    query_items = {
        "instance_id": "InstanceId",
        "iam_role": "IamInstanceProfile.Arn",
        "state": "State.Name",
        "tags": "Tags",
        "ip": "PublicIpAddress",
    }
    query = ",".join([f"{key}:{value}" for key, value in query_items.items()])
    instance_groups: List[List[Any]] = run_aws_command(
        [
            "aws",
            "ec2",
            "describe-instances",
            "--region",
            "us-west-2",
            "--filters",
            "Name=tag:ray-node-type,Values=*",
            "--query",
            f"Reservations[*].Instances[*].{{{query}}}",
        ]
    )
    return [
        InstanceInformation.from_dict(instance)
        for instance_group in instance_groups
        for instance in instance_group
    ]


def ssh_helper(
    final_config: dict,
    identity_file: Path,
    additional_port_forwards: list[int] = [],
) -> subprocess.Popen[str]:
    process = subprocess.Popen(
        ssh_command(
            ip=get_ip(final_config),
            user=final_config["ssh-user"],
            pub_key=identity_file,
            additional_port_forwards=additional_port_forwards,
        ),
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        text=True,
    )
    if process.returncode and process.returncode != 0:
        raise click.ClickException(
            f"Failed to attach to the remote server. Return code: {process.returncode}"
        )
    else:
        assert process.stdout
        if process.stdout.readable():
            if text := process.stdout.read():
                print(text)
        else:
            raise click.ClickException(
                "Unable to establish a connection to the remote server."
            )
    return process


def list_helper() -> dict[str, List[InstanceInformation]]:
    instance_infos = _parse_describe_instances_query()
    state_map = {}
    for instance_info in instance_infos:
        if instance_info.state not in state_map:
            state_map[instance_info.state] = []
        state_map[instance_info.state].append(instance_info)
    return state_map


def get_region(final_config: dict) -> str:
    if "provider" not in final_config:
        raise click.UsageError("The provider field is required in the configuration.")
    if "region" not in final_config["provider"]:
        raise click.UsageError("The region field is required in the provider field.")
    return final_config["provider"]["region"]


def get_instance_id(final_config: dict) -> str:
    name = final_config["cluster_name"]
    state_map = list_helper()
    if "running" not in state_map:
        raise click.UsageError(
            f"The cluster {name} is not running; cannot connect to it."
        )
    if len(state_map["running"]) == 0:
        raise click.UsageError(
            f"The cluster {name} is not running; cannot connect to it."
        )
    # for _, instance_id, node_type, _ in state_map["running"]:
    for instance_info in state_map["running"]:
        if instance_info.get_tag("ray-node-type") == "head":
            return instance_info.instance_id
    raise click.UsageError("Could not find the head node's Instance-Id.")


def query_for_public_keypair(final_config: dict) -> Optional[str]:
    region = get_region(final_config)
    instance_id = get_instance_id(final_config)
    keys: List[List[Any]] = run_aws_command(
        [
            "aws",
            "ec2",
            "describe-instances",
            "--region",
            region,
            "--instance-ids",
            instance_id,
            "--query",
            "Reservations[*].Instances[*].KeyName",
        ],
    )
    assert len(keys) == 1
    instance_keys = keys[0]
    num_of_keys = len(instance_keys)
    if num_of_keys == 0:
        return None
    elif num_of_keys == 1:
        return instance_keys[0]
    else:
        raise click.ClickException("This instance has multiple public key-pairs.")


def detect_keypair(final_config: dict) -> Path:
    if public_keypair_name := query_for_public_keypair(final_config):
        path = Path("~").expanduser() / ".ssh" / f"{public_keypair_name}.pem"
        if path.exists():
            return path
        else:
            raise click.ClickException(
                f"The public keypair name of the head node is called {public_keypair_name}, but no private keypair of that same name was found in the ~/.ssh directory; please re-run this command and manually pass in the path to the private keypair using the `-i <PATH_TO_KEY_PAIR>` flag."
            )
    else:
        raise click.UsageError(
            "Could not detect keypair; please manually specify one by using the `-i <PATH_TO_KEY_PAIR>` flag."
        )


def get_ip(final_config: dict):
    name = final_config["cluster_name"]
    region = get_region(final_config)
    instance_groups: List[List[Any]] = run_aws_command(
        [
            "aws",
            "ec2",
            "describe-instances",
            "--region",
            region,
            "--filters",
            "Name=tag:ray-node-type,Values=*",
            "--query",
            "Reservations[*].Instances[*].{State:State.Name,Tags:Tags,Ip:PublicIpAddress}",
        ],
    )
    state_to_ips_mapping = find_ip(instance_groups, name)
    if "running" not in state_to_ips_mapping:
        raise click.UsageError(
            f"The cluster {name} is not running; cannot connect to it."
        )
    assert len(state_to_ips_mapping["running"]) <= 1
    if state_to_ips_mapping["running"]:
        ip = state_to_ips_mapping["running"][0]
    else:
        raise click.UsageError(
            f"The cluster {name} is not running; cannot connect to it."
        )
    if not ip:
        raise click.UsageError(
            f"The cluster {name} does not have a public IP address available."
        )
    return ip


def run_aws_command(args: list[str]) -> Any:
    result = subprocess.run(args, capture_output=True, text=True)
    if result.returncode != 0:
        if "Token has expired" in result.stderr:
            raise click.UsageError(
                "AWS token has expired. Please run `aws login`, `aws sso login`, or some other command to refresh it."
            )
    if result.stdout == "":
        raise click.UsageError(
            f"Failed to parse AWS command into json (empty response string)"
        )
    try:
        return json.loads(result.stdout)
    except json.JSONDecodeError as e:
        raise click.UsageError(f"Failed to parse AWS command output: {result.stdout}")


def find_ip(
    instance_groups: List[List[Any]], name: str
) -> dict[str, list[Optional[str]]]:
    ip = None
    state_to_ips_mapping: dict[str, list[Optional[str]]] = {}

    def insert(state: str, ip: Optional[str]):
        if state in state_to_ips_mapping:
            state_to_ips_mapping[state].append(ip)
        else:
            state_to_ips_mapping[state] = [ip]

    for instance_group in instance_groups:
        for instance in instance_group:
            is_head = False
            cluster_name = None
            state = instance["State"]
            for tag in instance["Tags"]:
                if tag["Key"] == "ray-cluster-name":
                    cluster_name = tag["Value"]
                elif tag["Key"] == "ray-node-type":
                    is_head = tag["Value"] == "head"
            if is_head and cluster_name == name:
                insert(state, instance["Ip"])

    if not state_to_ips_mapping:
        raise click.UsageError(
            f"The IP of the cluster with the name '{name}' not found."
        )

    return state_to_ips_mapping


def ssh_command(
    ip: str,
    user: str | None,
    pub_key: Optional[Path] = None,
    additional_port_forwards: list[int] = [],
) -> list[str]:
    additional_port_forward_args = [
        arg
        for args in map(
            lambda pf: ["-L", f"{pf}:localhost:{pf}"], additional_port_forwards
        )
        for arg in args
    ]
    return (
        [
            "ssh",
            "-N",
            "-L",
            "8265:localhost:8265",
        ]
        + additional_port_forward_args
        + (["-i", str(pub_key)] if pub_key else [])
        + [
            f"{user}@{ip}" if user else f"ec2-user@{ip}",
        ]
    )


async def print_logs(logs):
    async for lines in logs:
        print(lines, end="")


async def wait_on_job(logs):
    await asyncio.wait_for(print_logs(logs), timeout=None)
